"""PeakRDL Python plug-in."""

__authors__ = ["Marek Pikuła <marek.pikula at embevity.com>"]

from typing import TYPE_CHECKING

from .exporter import PythonExporter

if TYPE_CHECKING:
    import argparse
    from typing import Union

    from systemrdl.node import AddrmapNode, RootNode  # type: ignore


class Exporter:  # pylint: disable=too-few-public-methods
    """PeakRDL Python exporter plug-in."""

    short_desc = "Export the register model to Python"

    def do_export(
        self, top_node: "Union[AddrmapNode, RootNode]", options: "argparse.Namespace"
    ):
        """Perform the export of SystemRDL node to Python.

        Arguments:
            top_node -- top node to export.
            options -- argparse options from the `peakrdl` tool.
        """
        PythonExporter().export(
            top_node,
            options.output,
            input_files=options.input_files,
            rename=options.inst_name,
        )



"""Reister interface simple test and demonstration of the package."""

from enum import IntEnum

from .access import FieldAccess, RegAccess
from .regif import DummyRegIf
from .spec import FieldNodeSpec, RegNodeSpec


class TestEnum(IntEnum):
    """Test enumaration type to check type casting for field access."""

    VALUE_0 = 0
    VALUE_1 = 1
    VALUE_2 = 2
    VALUE_4 = 4


class TestReg(RegAccess):
    """Test register."""

    _spec = RegNodeSpec(
        inst_name="test_reg",
        type_name="int",
        orig_type_name="int",
        external=False,
        raw_address_offset=0,
        address_offset=0,
        raw_absolute_address=0,
        absolute_address=0,
        size=1,
        total_size=1,
        is_array=False,
        array_dimensions=None,
        array_stride=None,
        is_virtual=False,
        has_sw_writable=True,
        has_sw_readable=True,
        has_hw_writable=True,
        has_hw_readable=True,
        is_interrupt_reg=False,
        is_halt_reg=False,
    )

    test_field = FieldAccess(
        FieldNodeSpec(
            inst_name="test_field",
            type_name="TestEnum",
            orig_type_name="TestEnum",
            external=False,
            width=3,
            msb=12,
            lsb=10,
            high=12,
            low=10,
            is_virtual=False,
            is_volatile=False,
            is_sw_writable=True,
            is_sw_readable=True,
            is_hw_writable=True,
            is_hw_readable=True,
            implements_storage=True,
            is_up_counter=False,
            is_down_counter=False,
            encode="TestEnum",
        ),
        TestEnum,
    )


if __name__ == "__main__":
    test_reg = TestReg(DummyRegIf(4 * 8, range(0, 0x1000), 0))

    # Simple read using TestEnum.
    print(test_reg.test_field)
    assert test_reg.test_field == TestEnum.VALUE_0

    # Simple write/read using TestEnum.
    test_reg.test_field = TestEnum.VALUE_1
    print(test_reg.test_field)
    assert test_reg.test_field == TestEnum.VALUE_1

    # Simple write/read using integer as written value.
    test_reg.test_field = 2  # type: ignore
    print(test_reg.test_field)
    assert test_reg.test_field == TestEnum.VALUE_2

    # Passing wrong value to the field.
    try:
        test_reg.test_field = 3  # type: ignore
    except ValueError:
        print("TestEnum cast error correctly caught.")



"""PeakRDL Python exporter."""

__authors__ = ["Marek Pikuła <marek.pikula at embevity.com>"]

import random
import string
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, OrderedDict, Set, Tuple, Union

from systemrdl.messages import MessageHandler  # type: ignore
from systemrdl.node import (  # type: ignore
    AddrmapNode,
    FieldNode,
    Node,
    RegfileNode,
    RegNode,
    RootNode,
)
from systemrdl.rdltypes.simple_enum import SimpleEnum  # type: ignore

from peakrdl_python_simple.regif.spec import (
    AddressableNodeSpec,
    AddrmapNodeSpec,
    FieldNodeSpec,
    NodeSpec,
    RegfileNodeSpec,
    RegNodeSpec,
)


class PythonExporter:  # pylint: disable=too-few-public-methods
    """PeakRDL Python exporter main class."""

    @dataclass
    class GenStageOutput:
        """Generation stage output."""

        node: Node
        """Node on which generation has been performed."""

        type_name: str
        """Name of Python type of this node."""

        spec: NodeSpec
        """Specification of the node."""

        generated_code: str
        """Python code generated during this stage."""

    def __init__(self):
        """Initialize the exporter."""
        # List of existing types to prevent duplication.
        self._existing_types: Set[str] = set()
        self._existing_enums: Set[str] = set()

    def export(
        self,
        node: Union[AddrmapNode, RootNode],
        output_path: str,
        input_files: Optional[List[str]] = None,
        rename: Optional[str] = None,
    ):
        """Export the `node` to generated Python interface file.

        Arguments:
            node -- node to export.
            input_files -- list of input files.
            output_path -- path to the exported file.
            rename -- name to rename the top-level to.
        """
        # Each export should start fresh.
        self._existing_types.clear()
        self._existing_enums.clear()

        # Get the top node.
        top = node.top if isinstance(node, RootNode) else node
        top_name = rename if rename is not None else node.inst_name

        generated_from = self._indent(1) + top_name
        if input_files is not None:
            generated_from = self._indent(1) + ("\n" + self._indent(1)).join(
                f for f in input_files
            )

        # Ensure proper format of the output path and that the directory exists.
        if not output_path.endswith(".py"):
            raise ValueError("The output file is not Python file.")
        Path(output_path).parent.mkdir(parents=True, exist_ok=True)

        # Generate the file.
        with open(output_path, "w", encoding="UTF-8") as output:
            output.write(
                (
                    '"""Python abstraction for SystemRDL register description.\n\n'
                    f"Don't override. Generated from:\n{generated_from}\n"
                    '"""\n\n'
                    "from enum import IntEnum\n\n"
                    "from peakrdl_python_simple.regif import access, spec\n"
                )
                + self._add_addrmap_regfile(
                    top, node.env.msg, is_top=True
                ).generated_code
            )

    INDENT_SPACES: int = 4

    @staticmethod
    def _indent(level: int = 0):
        return " " * PythonExporter.INDENT_SPACES * level

    def _generate_docstring(
        self,
        name: Optional[str],
        desc: Optional[str],
        indent_level: int = 1,
        add_final_newline: bool = True,
    ) -> str:
        """Generate docstring basing on the Node properties.

        Arguments:
            name -- name parameter from RDL.
            desc -- description parameter from RDL.

        Keyword Arguments:
            indent_level -- level of indentation.
            add_final_newline -- whether to add endline at the end or not of
                the docstring. If name and desc are empty the newline is not
                generated either way.

        Returns:
            Generated docstring.
        """
        indent_str = "\n" + self._indent(indent_level)
        name = "" if name is None else name.replace("\n", indent_str)
        desc = "" if desc is None else desc.replace("\n", indent_str)

        start = indent_str + '"""'
        end = '"""' + ("\n" if add_final_newline else "")
        if name == "" and desc == "":
            return ""
        if name != "" and desc == "":
            return start + name + end
        if name == "" and desc != "":
            return start + desc + end
        return start + name + "\n" + indent_str + desc + indent_str + end

    def _format_member(
        self, member: GenStageOutput, indent_level: int = 1, last: bool = False
    ) -> str:
        """Format class member.

        Arguments:
            member -- member definition to create the definition for.

        Keyword Arguments:
            indent_level -- level of indentation.
            last -- the member is the last in the class declaration.

        Returns:
            Generated member with docstring if applicable.
        """
        # Generate array suffix if applicable.
        array_suffix: List[str] = []
        if isinstance(member.spec, AddressableNodeSpec) and member.spec.is_array:
            assert (
                member.spec.array_dimensions is not None
                and member.spec.array_stride is not None
            )
            cur_index = (
                member.spec.absolute_address - member.spec.raw_absolute_address
            ) // member.spec.array_stride
            for dimension in reversed(member.spec.array_dimensions):
                array_suffix.append(f"_{cur_index % dimension}")
                cur_index //= dimension

        return (
            self._indent(indent_level)
            + f"{member.node.inst_name}{''.join(reversed(array_suffix))} = {member.type_name}("
            + f"specification=spec.{repr(member.spec)}"
            + (
                f", field_type={member.spec.encode}"
                if isinstance(member.spec, FieldNodeSpec)
                else ""
            )
            + ")"
            + self._generate_docstring(
                member.node.get_property("name", default=None),
                member.node.get_property("desc", default=None),
                indent_level,
                not last,
            )
        )

    @staticmethod
    def _to_pascal_case(unknown_str: str, check_suffix: Optional[str] = None):
        """Convert arbitrary string to PascalCase.

        Used for generating class names.

        Arguments:
            unknown_str -- arbitrary string.
            check_suffix -- suffix to check if exists in the string. If not, it's added.

        Returns:
            (Hopefully) PascalCase string.
        """
        out = unknown_str
        if "_" in unknown_str:
            out = "".join(
                s[0].upper() + (s[1:] if len(s) > 1 else "")
                for s in unknown_str.split("_")
            )
        if check_suffix is not None and not unknown_str.lower().endswith(
            check_suffix.lower()
        ):
            out += check_suffix
        return out[0].upper() + (out[1:] if len(out) > 1 else "")

    def _format_class(  # pylint: disable=too-many-arguments
        self,
        node: Node,
        spec: Optional[NodeSpec] = None,
        member_list: Optional[List[GenStageOutput]] = None,
        indent_level: int = 0,
        check_if_exists: bool = True,
    ) -> Tuple[str, str]:
        """Generage class definition.

        Arguments:
            node -- node to generate the class for.

        Keyword Arguments:
            spec -- optional specification list for `_spec` member generation.
            member_list -- optional list of members. If none are present `pass`
                is added.
            indent_level -- level of indentation for the class definition.
            check_if_exists -- check if the type exists and don't generate code
                if indeed it exists. Uses `self._existing_types` list.

        Returns:
            Class name and generated code tuple.
        """
        node_type = node.__class__.__name__.replace("Node", "")
        type_name = self._to_pascal_case(
            node.type_name + node_type
            if node.type_name is not None
            else "".join(random.choice(string.ascii_lowercase) for _ in range(16))
        )
        if check_if_exists:
            if type_name in self._existing_types:
                return type_name, ""
            self._existing_types.add(type_name)

        gen = (
            "\n\n"
            + self._indent(indent_level)
            + f"class {type_name}(access.{node_type}Access):"
            + self._generate_docstring(
                node.get_property("name", default=None),
                node.get_property("desc", default=None),
                indent_level + 1,
                True,
            )
            + "\n"
        )
        if spec is not None:
            gen += self._indent(indent_level + 1) + f"_spec = spec.{repr(spec)}\n"
        if member_list is not None and len(member_list) > 0:
            gen += "\n".join(
                self._format_member(member, indent_level + 1, i == len(member_list) - 1)
                for i, member in enumerate(member_list)
            )
        else:
            gen += self._indent(indent_level + 1) + "pass"

        return type_name, gen + "\n"

    def _add_addrmap_regfile(
        self,
        node: Union[AddrmapNode, RegfileNode],
        msg: MessageHandler,
        is_top: bool = False,
    ) -> GenStageOutput:
        """Generate addrmap or regfile.

        Arguments:
            node -- RegfileNode or AddrmapNode.
            msg -- message handler from top-level.

        Keyword Arguments:
            is_top -- if the current not is the top node. If True the
                specification is embedded as class member.

        Returns:
            Generated addrmap output.
        """
        members: List[PythonExporter.GenStageOutput] = []
        gen: str = ""
        for child in node.children(unroll=True, skip_not_present=False):
            if isinstance(child, (AddrmapNode, RegfileNode)):
                output = self._add_addrmap_regfile(child, msg)
                gen += output.generated_code
                members.append(output)
            elif isinstance(child, RegNode):
                output = self._add_reg(child, msg)
                gen += output.generated_code
                members.append(output)
            else:
                msg.warning(
                    f"Unsupported type of node ({child.type_name}) for {child.inst_name}."
                )

        spec_type = (
            RegfileNodeSpec if isinstance(node, RegfileNode) else AddrmapNodeSpec
        )
        spec = spec_type(
            node.inst_name,
            node.type_name,
            node.orig_type_name,
            node.external,
            node.raw_address_offset,
            node.address_offset,
            node.raw_absolute_address,
            node.absolute_address,
            node.size,
            node.total_size,
            node.is_array,
            node.array_dimensions,
            node.array_stride,
        )
        type_name, gen_node = self._format_class(
            node, spec if is_top else None, members
        )
        return PythonExporter.GenStageOutput(node, type_name, spec, gen + gen_node)

    def _add_reg(self, node: RegNode, msg: MessageHandler) -> GenStageOutput:
        """Generate register.

        Arguments:
            node -- RegNode.
            msg -- message handler from top-level.

        Returns:
            Generated register output.
        """
        gen: str = ""
        members: List[PythonExporter.GenStageOutput] = []
        for field in node.fields(skip_not_present=True):
            output = self._add_field(field, msg)
            gen += output.generated_code
            members.append(output)

        type_name, gen_node = self._format_class(node, member_list=members)
        return PythonExporter.GenStageOutput(
            node,
            type_name,
            RegNodeSpec(
                node.inst_name,
                node.type_name,
                node.orig_type_name,
                node.external,
                node.raw_address_offset,
                node.address_offset,
                node.raw_absolute_address,
                node.absolute_address,
                node.size,
                node.total_size,
                node.is_array,
                node.array_dimensions,
                node.array_stride,
                node.is_virtual,
                node.has_sw_writable,
                node.has_sw_readable,
                node.has_hw_writable,
                node.has_hw_readable,
                node.is_interrupt_reg,
                node.is_halt_reg,
            ),
            gen + gen_node,
        )

    def _add_field(
        self,
        node: FieldNode,
        msg: MessageHandler,
    ) -> GenStageOutput:
        """Generate field.

        Arguments:
            node -- FieldNode.
            msg -- message handler from top-level.

        Returns:
            Generated field output.
        """
        encode_type: str = "int" if node.width > 1 else "bool"
        gen = ""

        enum: Optional[SimpleEnum] = node.get_property("encode", default=None)
        if enum is not None:
            encode_type, gen = self._add_enum(enum, msg)
        return PythonExporter.GenStageOutput(
            node,
            "access.FieldAccess",
            FieldNodeSpec(
                node.inst_name,
                node.type_name,
                node.orig_type_name,
                node.external,
                node.width,
                node.msb,
                node.lsb,
                node.high,
                node.low,
                node.is_virtual,
                node.is_volatile,
                node.is_sw_writable,
                node.is_sw_readable,
                node.is_hw_writable,
                node.is_hw_readable,
                node.implements_storage,
                node.is_up_counter,
                node.is_down_counter,
                encode_type,
            ),
            gen,
        )

    def _add_enum(
        self,
        enum: SimpleEnum,
        msg: MessageHandler,  # pylint: disable=unused-argument
    ) -> Tuple[str, str]:
        """Add enum class.

        Arguments:
            enum -- SystemRDL enum definition.
            msg -- message handler from top-level.

        Returns:
            Generated Enum class name and generated code tuple.
        """
        class_name = self._to_pascal_case(enum.__name__, "Enum")  # type:ignore

        gen = ""
        if class_name not in self._existing_enums:
            self._existing_enums.add(class_name)
            members: OrderedDict[str, SimpleEnum] = enum.__members__  # type:ignore
            gen += (
                f"\n\nclass {class_name}(IntEnum):\n"
                + self._indent(1)
                + ("\n" + self._indent(1)).join(
                    self._format_enum_member(
                        name, enum_item, i == len(members.items()) - 1
                    )
                    for i, (name, enum_item) in enumerate(members.items())
                )
                + "\n"
            )

        return class_name, gen

    def _format_enum_member(self, name: str, enum_member: SimpleEnum, last: bool):
        """Format SystemRDL enum member.

        Arguments:
            name -- name of the enum.
            enum_member -- SimpleEnum from SystemRDL.
            last -- whether the current member is the last in the enum class.

        Returns:
            Generated Python enum member.
        """
        value: int = enum_member.value  # type: ignore
        rdl_name: Optional[
            str
        ] = enum_member._rdl_name_  # type:ignore # pylint: disable=protected-access
        rdl_desc: Optional[
            str
        ] = enum_member._rdl_desc_  # type:ignore # pylint: disable=protected-access

        return f"{name} = {value}" + self._generate_docstring(
            rdl_name,
            rdl_desc,
            add_final_newline=not last,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['peakrdl_python_simple', 'peakrdl_python_simple.regif']

package_data = \
{'': ['*']}

install_requires = \
['systemrdl-compiler>=1.23.0,<2.0.0']

extras_require = \
{'cli': ['peakrdl>=0.3.0,<0.4.0']}

entry_points = \
{'peakrdl.exporters': ['python-simple = '
                       'peakrdl_python_simple.__peakrdl__:Exporter']}

setup_kwargs = {
    'name': 'peakrdl-python-simple',
    'version': '0.2.0',
    'description': 'Export Python description from the systemrdl-compiler register model',
    'long_description': '[![Documentation Status](https://readthedocs.org/projects/peakrdl-python-simple/badge/?version=latest)](http://peakrdl-python-simple.readthedocs.io)\n[![build](https://github.com/MarekPikula/PeakRDL-Python-simple/workflows/build/badge.svg)](https://github.com/MarekPikula/PeakRDL-Python-simple/actions?query=workflow%3Abuild+branch%3Amain)\n[![Coverage Status](https://coveralls.io/repos/github/MarekPikula/PeakRDL-Python-simple/badge.svg?branch=main)](https://coveralls.io/github/MarekPikula/PeakRDL-Python-simple?branch=main)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/peakrdl-python-simple.svg)](https://pypi.org/project/peakrdl-python-simple)\n\n# PeakRDL-Python\n\nThis package implements Python register abstraction layer export for the\nPeakRDL toolchain.\n\n- **Export:** Convert compiled SystemRDL input into Python register interface.\n\nFor the command line tool, see the [PeakRDL\nproject](https://peakrdl.readthedocs.io).\n\n## Usage\n\nPeakRDL project provides a standard CLI interface. It can be installed directly\nvia pip or by installing this package with `cli` extra:\n\n    $ pip install peakrdl-python-simple[cli]\n\nThen this package can be used with the following command:\n\n    $ peakrdl python-simple input_file.rdl -o output_interface.py\n\n## Documentation\n\nSee the [PeakRDL-Python-simple\nDocumentation](http://peakrdl-python-simple.readthedocs.io) for more details.\n',
    'author': 'Marek Pikuła',
    'author_email': 'marek.pikula@embevity.com',
    'maintainer': 'Marek Pikuła',
    'maintainer_email': 'marek.pikula@embevity.com',
    'url': 'https://github.com/MarekPikula/PeakRDL-Python-simple',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)



"""Node specification dataclasses.

All are based on available properties of their systemrdl.node.* counterparts.
"""

__authors__ = ["Marek Pikuła <marek.pikula at embevity.com>"]

from dataclasses import dataclass
from typing import List, Optional


@dataclass(frozen=True)
class NodeSpec:
    """Specification of systemrdl.node.Node."""

    # owning_addrmap: Optional["AddrmapNodeSpec"]
    inst_name: str
    type_name: Optional[str]
    orig_type_name: Optional[str]
    external: bool
    # cpuif_reset: Optional["SignalNode"]


@dataclass(frozen=True)
class AddressableNodeSpec(NodeSpec):  # pylint: disable=too-many-instance-attributes
    """Specification of systemrdl.node.AddressableNode."""

    raw_address_offset: int
    address_offset: int
    raw_absolute_address: int
    absolute_address: int
    size: int
    total_size: int
    is_array: bool
    array_dimensions: Optional[List[int]]
    array_stride: Optional[int]


@dataclass(frozen=True)
class VectorNodeSpec(NodeSpec):
    """Specification of systemrdl.node.VectorNode."""

    width: int
    msb: int
    lsb: int
    high: int
    low: int


@dataclass(frozen=True)
class RootNodeSpec(NodeSpec):
    """Specification of systemrdl.node.RootNode."""


@dataclass(frozen=True)
class SignalNodeSpec(VectorNodeSpec):
    """Specification of systemrdl.node.SignalNode."""


@dataclass(frozen=True)
class FieldNodeSpec(VectorNodeSpec):  # pylint: disable=too-many-instance-attributes
    """Specification of systemrdl.node.FieldNode."""

    is_virtual: bool
    is_volatile: bool
    is_sw_writable: bool
    is_sw_readable: bool
    is_hw_writable: bool
    is_hw_readable: bool
    implements_storage: bool
    is_up_counter: bool
    is_down_counter: bool
    # Aliases not supported:
    #   is_alias: bool
    #   alias_primary: "FieldNodeSpec"
    #   has_aliases: bool
    #   aliases: Iterator["FieldNodeSpec"]
    encode: str


@dataclass(frozen=True)
class RegNodeSpec(AddressableNodeSpec):
    """Specification of systemrdl.node.RegNode."""

    is_virtual: bool
    has_sw_writable: bool
    has_sw_readable: bool
    has_hw_writable: bool
    has_hw_readable: bool
    is_interrupt_reg: bool
    is_halt_reg: bool
    # Aliases not supported:
    #   is_alias: bool
    #   alias_primaty: "RegNodeSpec"
    #   has_aliases: bool
    #   aliases: Iterator["RegNodeSpec"]


@dataclass(frozen=True)
class RegfileNodeSpec(AddressableNodeSpec):
    """Specification of systemrdl.node.RegfileNode."""


@dataclass(frozen=True)
class AddrmapNodeSpec(AddressableNodeSpec):
    """Specification of systemrdl.node.AddrmapNode."""


@dataclass(frozen=True)
class MemNodeSpec(AddressableNodeSpec):
    """Specification of systemrdl.node.MemNode."""

    is_sw_writable: bool
    is_sw_readable: bool



"""Register interface abstraction."""

__authors__ = ["Marek Pikuła <marek.pikula at embevity.com>"]

from abc import ABC, abstractmethod
from typing import Dict, Optional


class RegisterInterface(ABC):
    """Register interface abstraction.

    A basic register interface requires overriding `get()` and `set()`
    functions, depending on underlying hardware configuration.

    As an example implementation `DummyRegIf` is available.
    """

    def __init__(self, data_width: int, address_bounds: Optional[range] = None):
        """Initialize register interface abstraction.

        Arguments:
            data_width -- width of data in bits, should be divisible by 8.

        Keyword Arguments:
            address_bounds -- address range, which is allowed by this register
                interface. If not defined, addresses are not validated if they
                are in range.

        Raises:
            ValueError: raised if sanity check on the arguments doesn't pass.
        """
        if not 64 >= data_width > 0:
            raise ValueError("Unsupported register width.")
        if data_width % 8 != 0:
            raise ValueError("Data width should be divisible by 8 (a byte).")
        if address_bounds is not None:
            if address_bounds.start < 0:
                raise ValueError("Address bounds need to be positive.")
            if address_bounds.start > address_bounds.stop:
                raise ValueError("Address bounds need to be incremental.")

        self._data_width = data_width
        self._address_bounds = address_bounds

    def _sanitize_field_args(
        self,
        reg_address: int,
        field_pos: Optional[int] = None,
        field_width: Optional[int] = None,
        value: Optional[int] = None,
    ):
        """General argument sanitizer used both for register and field access.

        It's assumed that it's field if `field_pos` and `field_width_bits` are
        defined.

        Arguments:
            reg_address -- register address.

        Keyword Arguments:
            field_pos -- field position.
            field_width -- field width in bits.
            value -- value of register or field.

        Raises:
            ValueError: some inconsistency has been found.
        """
        if reg_address < 0:
            raise ValueError(f"Register address {reg_address} should be positive.")

        if self._address_bounds is not None and reg_address not in self._address_bounds:
            raise ValueError(
                f"Register address 0x{reg_address:X} not in register interface allowed range."
            )

        if field_pos is not None and field_pos < 0:
            raise ValueError(f"Field position ({field_pos}) should be positive.")

        if field_width is not None and field_width > self.data_width:
            raise ValueError(
                f"Field width ({field_width}) should not be bigger than"
                f"register width ({self.data_width})."
            )

        # If field use field_width otherwise use register data_width.
        bits = (
            field_width
            if field_pos is not None and field_width is not None
            else self.data_width
        )
        if value is not None and value & ((1 << bits) - 1) != value:
            raise ValueError(
                f"Register/field value (0x{value:X}) wider than "
                f"register/field width ({bits})."
            )

    @property
    def data_width(self):
        """Get register data width."""
        return self._data_width

    @abstractmethod
    def get(self, reg_address: int) -> int:
        """Read register value abstraction.

        `super().get(reg_address)` should be called to validate aruments.

        Arguments:
            reg_address -- absolute address of register to read.

        Returns:
            Data from the register.
        """
        self._sanitize_field_args(reg_address)
        return 0

    @abstractmethod
    def set(self, reg_address: int, value: int) -> None:
        """Write register value abstraction.

        `super().set(reg_address, value)` should be called to validate aruments.

        Arguments:
            reg_address -- absolute address of register to write to.
            value -- value to write to the register.
        """
        self._sanitize_field_args(reg_address, value=value)

    def get_field(self, reg_address: int, field_pos: int, field_width: int) -> int:
        """Read register field abstraction.

        It sanitized arguments to ensure nothing is out of bounds.

        The default implementation can be overloaded, but
        `_sanitize_field_args()` should be called in the
        overloaded implementation.

        Arguments:
            reg_address -- absolute address of register to write to.
            field_pos -- field position in the register (counting from LSB).
            field_width -- width of the field in bits.

        Returns:
            Value in given field in given register.
        """
        self._sanitize_field_args(reg_address, field_pos, field_width)
        return (self.get(reg_address) >> field_pos) & ((1 << field_width) - 1)

    def set_field(
        self, reg_address: int, field_pos: int, field_width: int, value: int
    ) -> None:
        """Write register field abstraction.

        It sanitized arguments to ensure nothing is out of bounds.

        The default implementation can be overloaded, but
        `_sanitize_field_args()` should be called in the
        overloaded implementation.

        Arguments:
            reg_address -- absolute address of register to write to.
            field_pos -- field position in the register (counting from LSB).
            field_width -- width of the field in bits.
            value -- new value of the field.
        """
        self._sanitize_field_args(reg_address, field_pos, field_width, value)

        field_negative_mask = ((1 << self.data_width) - 1) ^ (
            ((1 << field_width) - 1) << field_pos
        )
        self.set(
            reg_address,
            self.get(reg_address) & field_negative_mask | (value << field_pos),
        )


class DummyRegIf(RegisterInterface):
    """Example implementation of RegisterInterface.

    It stores the register file in memory as dictionary.
    """

    def __init__(
        self, data_width: int, address_bounds: Optional[range], reset_value: int = 0
    ):
        """Initialize the dummy register interface.

        Arguments:
            data_width -- width of data in bits, should be divisible by 8.

        Keyword Arguments:
            address_bounds -- address range, which is allowed by this register
                interface. If not defined, addresses are not validated if they
                are in range.
            reset_value -- default value present in the register on "reset".
        """
        super().__init__(data_width, address_bounds)
        self._values: Dict[int, int] = {}
        self._reset_value = reset_value

    def get(self, reg_address: int) -> int:
        """Get value from register.

        Arguments:
            reg_address -- absolute register address.

        Returns:
            Register value.
        """
        super().get(reg_address)
        if reg_address not in self._values:
            return self._reset_value
        return self._values[reg_address]

    def set(self, reg_address: int, value: int):
        """Set register value.

        Arguments:
            reg_address -- absolute register address.
            value -- value to write to the register.
        """
        super().set(reg_address, value)
        self._values[reg_address] = value
